﻿using System;
using System.Collections.Generic;
using System.Linq;

class Interpreter
{
    List<Rule> AllRules;
    List<Rule> RulesToRun = new List<Rule>();
	List<Rule> RecentAssignments = new List<Rule>();
	List<Rule> RecentRanRules = new List<Rule>();
	List<Rule> During = new List<Rule>();
	Rule ThePlayersCommand;
	bool RulesDebuggingCommand = false;
	readonly AbstractSyntaxTree Yes = new AbstractSyntaxTree("true", ValueTypes.boolean);
	readonly AbstractSyntaxTree No = new AbstractSyntaxTree("false", ValueTypes.boolean);

    public void RunProgram(List<Rule> allRules)
    {
        AllRules = allRules;
		ThePlayersCommand = AllRules.Find(r => r.Atom == "ThePlayersCommand");
		Rule MainLoop = allRules.Find(r => r.Atom == "MainLoop");
		
		Console.WriteLine("\n=== Start Program ==========\n");
        RulesToRun.Add(AllRules.Find(r => r.Atom == "ProgramStart"));
	    while (string.Compare(ThePlayersCommand.Code.Value, "quit", StringComparison.InvariantCultureIgnoreCase) != 0)
	    {
		    foreach (var rule in RulesToRun)
		    {
			    if (RulesDebuggingCommand) Console.WriteLine("[Rule " + rule.Atom + " applies.]");
			    During.Add(rule);
			    EvalCode(rule.Code);
				During.Remove(rule);
				RecentRanRules.Add(rule);
		    }
		    RulesToRun.Clear();
			foreach (var rule in AllRules)
				foreach (var modifiedRule in RecentAssignments)
					if (rule.Condition != null && rule.Condition.Contains(modifiedRule.Atom))
						if (EvalCode(rule.Condition).Value == "true")
						{
							RulesToRun.Add(rule);
							break;
						}
		    if (RulesToRun.Count == 0)
			    RulesToRun.Add(MainLoop);
		    RecentAssignments.Clear();
	    }
	    Console.WriteLine("\n=== End Program ===========\n");
    }

    AbstractSyntaxTree EvalCode(AbstractSyntaxTree node)
    {
        if (node == null) return null;
        if (node.IsOperand) return node;
		AbstractSyntaxTree left = EvalCode(node.Left);
		if (node.Value == "?")
		{
			left = TestRelation(left);
			if (left != Yes) return null;
	    }
        AbstractSyntaxTree right = EvalCode(node.Right);
	    switch (node.Value)
	    {
			case "?":
				InvokeFunction(right);
				return null;

			case "{":
			case ";":
			    InvokeFunction(left);
			    InvokeFunction(right);
			    return null;

		    case ":=": 
			    var rule = AllRules.FirstOrDefault(r => r.Atom == left.ToString());
			    if (rule == null)
			    {
				    rule = new Rule {Atom = left.ToString(), Code = right};
				    AllRules.Add(rule);
			    }
			    if (right.Value == "ReadKeyboard")
					rule.Code = new AbstractSyntaxTree(Console.ReadLine(), ValueTypes.text);
				else if (right.Value == "ReadKey")
					rule.Code = new AbstractSyntaxTree(Console.ReadKey().KeyChar.ToString(), ValueTypes.text);
			    else
				    rule.Code = right;
				RecentAssignments.Add(rule);
				return null;

			case "(":
			    return SubstituteParamsWithValues(node);

	    }
		// at this point, any items of the shape   foo(bar,blah)   are relations, not function-invocations. We test for existance and value.
	    left = TestRelation(left);
	    right = TestRelation(right);
		// these produce a value. Statements (above) do not produce values.
		switch (node.Value)
        {
			case "<":  return left.Value.CompareTo(right.Value) < 0 ? Yes : No;
			case "<=": return left.Value.CompareTo(right.Value) <= 0 ? Yes : No;
            case "<>":
			case "!=": return left.Value.CompareTo(right.Value) != 0 ? Yes : No;
			case ">":  return left.Value.CompareTo(right.Value) > 0 ? Yes : No;
			case ">=": return left.Value.CompareTo(right.Value) >= 0 ? Yes : No;
			case "==": return left.Value.CompareTo(right.Value) == 0 ? Yes : No;
            case "&":
			case "&&": return bool.Parse(left.Value) && bool.Parse(right.Value) ? Yes : No;
            case "|":
			case "||": return bool.Parse(left.Value) || bool.Parse(right.Value) ? Yes : No;
			case "+":  return new AbstractSyntaxTree((decimal.Parse(left.Value) + decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "-":  return new AbstractSyntaxTree((decimal.Parse(left.Value) - decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "*":  return new AbstractSyntaxTree((decimal.Parse(left.Value) * decimal.Parse(right.Value)).ToString(), ValueTypes.number);
			case "/":  return new AbstractSyntaxTree((decimal.Parse(left.Value) / decimal.Parse(right.Value)).ToString(), ValueTypes.number);
		}
        return node;
    }

	// when node.Value == "("
	AbstractSyntaxTree SubstituteParamsWithValues(AbstractSyntaxTree node)
	{
		if (node.Right == null || node.Right.Value == null)
			return node; // there are no parameters
		var retval = new AbstractSyntaxTree(node);
		int numUnboundParameters = 0;
		for (var subnode = retval.Right; subnode != null; subnode = subnode.Right)
		{
			if (subnode.Value == ",")
			{
				var parameter = AllRules.FirstOrDefault(r => r.Atom == subnode.Left.Value);
				if (parameter != null)
					subnode.Left.Value = EvalCode(parameter.Code).ToString();
				else
					numUnboundParameters++;
			}
			else
			{
				var parameter = AllRules.FirstOrDefault(r => r.Atom == subnode.Value);
				if (parameter != null)
					subnode.Value = EvalCode(parameter.Code).ToString();
				else
					numUnboundParameters++;
			}
			if (numUnboundParameters > 0)
				Console.Write("");
		}
		return retval;
	}

	// all parameters should be bound to arguments, but the relation-name needn't actually exist
	AbstractSyntaxTree TestRelation(AbstractSyntaxTree node)
	{
		if (node.ValueType == ValueTypes.code)
		{
			var rule = AllRules.Find(r => r.Atom == node.ToString());
			node = (rule != null) ? rule.Code ?? Yes : No;
		}
		return node;
	}

	// all parameters should be bound to arguments
	void InvokeFunction(AbstractSyntaxTree node)
    {
		if (node == null || node.Value == null)
			return;
		if (RulesDebuggingCommand && node.ValueType == ValueTypes.code)
			Console.WriteLine("[Rule " + node + " applies.]");
	    if (node.Value == "Parse")
	    {
		    string input = ThePlayersCommand.Code.Value;
		    var words = input.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
		    var verb = AllRules.Find(r => r.Atom == "verb");
		    var noun = AllRules.Find(r => r.Atom == "noun");
		    var prep = AllRules.Find(r => r.Atom == "prep");
		    var noun2 = AllRules.Find(r => r.Atom == "noun2");
			verb.Code.Value = (words.Length >= 1) ? words[0] : "";
			noun.Code.Value = (words.Length >= 2) ? words[1] : "";
			prep.Code.Value = (words.Length >= 3) ? words[2] : "";
			noun2.Code.Value = (words.Length >= 4) ? words[3] : "";
		    Rule referredObject;
			referredObject = AllRules.Find(r => r.Atom == verb.Code.Value);
			if (referredObject != null) verb.Code = referredObject.Code;
			referredObject = AllRules.Find(r => r.Atom == noun.Code.Value);
			if (referredObject != null) noun.Code = referredObject.Code;
			referredObject = AllRules.Find(r => r.Atom == prep.Code.Value);
			if (referredObject != null) prep.Code = referredObject.Code;
			referredObject = AllRules.Find(r => r.Atom == noun2.Code.Value);
			if (referredObject != null) noun2.Code = referredObject.Code;
			RecentAssignments.Add(verb);
			RecentAssignments.Add(noun);
			RecentAssignments.Add(prep);
			RecentAssignments.Add(noun2);
		    if (words.Length >= 1 && string.Compare(words[0], "rules", StringComparison.InvariantCultureIgnoreCase) == 0)
		    {
			    RulesDebuggingCommand = !RulesDebuggingCommand;
			    Console.WriteLine("[The RULES debugging command is now " + RulesDebuggingCommand + "]");
		    }
		    return;
	    }
	    if (node.ValueType == ValueTypes.code) // then run all rules with that name
	    {
		    var funcName = node.Value == "(" ? node.Left.Value : node.Value;
		    foreach (Rule invoked in AllRules.Where(r => r.Atom == funcName))
		    {
			    if (invoked == null || invoked.Code == null) continue;
				//During.Add(invoked);
				Do(invoked.Code);
				//During.Remove(invoked);
				//RecentRanRules.Add(invoked);
			}
	    }
	    else
		    Do(node);
    }

	private void Do(AbstractSyntaxTree node)
	{
		// ...and then depending on the value's type, print or execute it.
		if (node.ValueType == ValueTypes.code)
			EvalCode(node);
		else if (node.ValueType == ValueTypes.text)
			Console.Write(node.Value);
		else if (node.ValueType == ValueTypes.number)
			Console.Write(node.Value);
		else if (node.ValueType == ValueTypes.boolean)
			Console.Write(node.Value == "true" ? "yes" : "no");
		else
			throw new Exception("We never defined:  do " + node.Value + "... ");
	}
}
    